//
// Improved Harris Corner Detector (Color version)
//

#include	<math.h>
#include	"CHarris.h"
#include	"HarrisTools.h"

#define		Mark_Red		255

// Improved Harris corner detector w/o output image
int
CImpHarris(int *Im, int width, int height,
           int **up0, int **vp0, int *cnr_num,
           int Ns, double sigma, double rk,
           double s, int max_corner, double min_dist)
{
   double	r, d;
   double	Ix2, Iy2, Ixy;
   double       gx, gy;
   double	gxx, gyy, gxy;
   double	max_r, r_thd;
   int		corner_num;
   int		x, x0, xx, u0, v0;
   int		xp2, xp1, xm1, xm2, yp2, yp1, ym1, ym2;
   int		uoff, voff, offwidth, offheight;
   int		dblw;
   
   int		*up = new int [max_corner];
   int		*vp = new int [max_corner];
   
   double	*Gausw   = new double [(2*Ns+1) * (2*Ns+1)];
   float	*Rx  = new float [width * height];
   float	*Ry  = new float [width * height];
   float	*Gx  = new float [width * height];
   float	*Gy  = new float [width * height];
   float	*Bx  = new float [width * height];
   float	*By  = new float [width * height];
   float	*Rxx = new float [width * height];
   float	*Ryy = new float [width * height];
   float	*Rxy = new float [width * height];
   float	*Gxx = new float [width * height];
   float	*Gyy = new float [width * height];
   float	*Gxy = new float [width * height];
   float	*Bxx = new float [width * height];
   float	*Byy = new float [width * height];
   float	*Bxy = new float [width * height];
   float	*out  = new float [width * height];
   double	*dp = new double [max_corner];

   // memory allocation check
   if (!up || !vp || !dp || !out || !Gausw ||
       !Rx || !Ry || !Rxx || !Ryy || !Rxy ||
       !Gx || !Gy || !Gxx || !Gyy || !Gxy ||
       !Bx || !By || !Bxx || !Byy || !Bxy)
      return 0;

   (*up0) = up;
   (*vp0) = vp;

   make_Gauss_window(Gausw, Ns, sigma);

   dblw = width * 2;

   uoff = voff = 2;
   offwidth = width - 2;
   offheight = height - 2;
   
   for (v0 = voff; v0 < offheight; v0++)
   {
      x = (v0 * width) + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 xp2 = (x + 2) * 3;
	 xp1 = (x + 1) * 3;
	 xm1 = (x - 1) * 3;
	 xm2 = (x - 2) * 3;
	 yp2 = (x + dblw) * 3;
	 yp1 = (x + width) * 3;
	 ym1 = (x - width) * 3;
	 ym2 = (x - dblw) * 3;
	 Rx[x] = G2 * (Im[xp2] - Im[xm2]) + G1 * (Im[xp1] - Im[xm1]);
	 Ry[x] = G2 * (Im[yp2] - Im[ym2]) + G1 * (Im[yp1] - Im[ym1]);
	 Gx[x] = G2 * (Im[xp2+1] - Im[xm2+1]) + G1 * (Im[xp1+1] - Im[xm1+1]);
	 Gy[x] = G2 * (Im[yp2+1] - Im[ym2+1]) + G1 * (Im[yp1+1] - Im[ym1+1]);
	 Bx[x] = G2 * (Im[xp2+2] - Im[xm2+2]) + G1 * (Im[xp1+2] - Im[xm1+2]);
	 By[x] = G2 * (Im[yp2+2] - Im[ym2+2]) + G1 * (Im[yp1+2] - Im[ym1+2]);
      }
   }

   uoff += 2;
   voff += 2;
   offwidth -= 2;
   offheight -= 2;

   for (v0 = voff; v0 < offheight; v0++)
   {
      x = (v0 * width) + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 gx = Rx[x - 2] + Rx[x - 1] + Rx[x] + Rx[x + 1] + Rx[x + 2];
	 gy = Ry[x - dblw] + Ry[x - width] + Ry[x]
	    + Ry[x + width] + Ry[x + dblw];
	 Rxx[x] = gx * gx;
	 Ryy[x] = gy * gy;
	 Rxy[x] = gx * gy;
	 gx = Gx[x - 2] + Gx[x - 1] + Gx[x] + Gx[x + 1] + Gx[x + 2];
	 gy = Gy[x - dblw] + Gy[x - width] + Gy[x]
	    + Gy[x + width] + Gy[x + dblw];
	 Gxx[x] = gx * gx;
	 Gyy[x] = gy * gy;
	 Gxy[x] = gx * gy;
	 gx = Bx[x - 2] + Bx[x - 1] + Bx[x] + Bx[x + 1] + Bx[x + 2];
	 gy = By[x - dblw] + By[x - width] + By[x]
	    + By[x + width] + By[x + dblw];
	 Bxx[x] = gx * gx;
	 Byy[x] = gy * gy;
	 Bxy[x] = gx * gy;
      }
   }

   uoff += Ns;
   voff += Ns;
   offwidth -= Ns;
   offheight -= Ns;

   max_r = 0.0;
   
   for (v0 = voff; v0 < offheight; v0++)
   {
      x = v0 * width + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 gxx = gyy = gxy = 0.0;
	 xx = 0;
	 for (int vv = -Ns; vv <= Ns; vv++)
	 {
	    x0 = (v0 + vv) * width + u0;
	    for (int uu = -Ns; uu <= Ns; uu++)
	    {
	       gxx += Gausw[xx] * (Rxx[x0 + uu] + Gxx[x0 + uu] + Bxx[x0 + uu]);
	       gyy += Gausw[xx] * (Ryy[x0 + uu] + Gyy[x0 + uu] + Byy[x0 + uu]);
	       gxy += Gausw[xx] * (Rxy[x0 + uu] + Gxy[x0 + uu] + Bxy[x0 + uu]);
	       xx++;
	    }
	 }
	 
	 Ix2 = gxx * gxx;
	 Iy2 = gyy * gyy;
	 Ixy = gxy * gxy;
	 r = (Ix2 * Iy2 - sqr(Ixy)) - rk * sqr(Ix2 + Iy2);

	 if (r > max_r)
	    max_r = r;
	 
	 out[x] = r;
      }
   }

   r_thd = max_r * s;

   uoff += 2;
   voff += 2;
   offwidth -= 2;
   offheight -= 2;

   corner_num = 0;
   for (v0 = voff; v0 < offheight; v0++)
   {
      x = v0 * width + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 d = out[x];
	 if (d <= r_thd)  continue;

	 if (   check_ge_upperleft(d, out, x, width)
             && check_gt_lowerright(d, out, x, width))
	 {
	    corner_num = add_list(u0, v0, d, up, vp, dp,
				  corner_num, max_corner, min_dist);
	 }
      }
   }

   delete [] dp;
   delete [] out;
   delete [] Rxy;
   delete [] Ryy;
   delete [] Rxx;
   delete [] Ry;
   delete [] Rx;
   delete [] Gxy;
   delete [] Gyy;
   delete [] Gxx;
   delete [] Gy;
   delete [] Gx;
   delete [] Bxy;
   delete [] Byy;
   delete [] Bxx;
   delete [] By;
   delete [] Bx;
   delete [] Gausw;
   
   return (*cnr_num = corner_num);
}

// Harris Color corner detector
int
CImpHarris(int *Im, int *Imo, int width, int height,
           int **up0, int **vp0, int *cnr_num,
           int Ns, double sigma, double rk,
           double s, int max_corner, double min_dist)
{
   int		x, corner_num;

   // feature detection
   corner_num = CImpHarris(Im, width, height, up0, vp0, cnr_num,
                           Ns, sigma, rk, s, max_corner, min_dist);

   // copy image
   x = 0;
   for (int v = 0; v < height; v++)
      for (int u = 0; u < width; u++)
      {
	 Imo[x++] = Im[x];
	 Imo[x++] = Im[x];
	 Imo[x++] = Im[x];
      }

   // plot corner
   for (int i = 0; i < corner_num; i++)
   {
      x = (*vp0)[i] * width + (*up0)[i];
      Imo[(x - width - 1) * 3] = Mark_Red;
      Imo[(x - width) * 3]     = Mark_Red;
      Imo[(x - width + 1) * 3] = Mark_Red;
      Imo[(x - 1) * 3]         = Mark_Red;
      Imo[x * 3]               = 0;
      Imo[(x + 1) * 3]         = Mark_Red;
      Imo[(x + width - 1) * 3] = Mark_Red;
      Imo[(x + width) * 3]     = Mark_Red;
      Imo[(x + width + 1) * 3] = Mark_Red;
   }
   return corner_num;
}
