//
// Improved Harris Corner Detector
//

#include	<math.h>
#include	"Harris.h"
#include	"HarrisTools.h"

#define		MaxGray		255
#define         G1              0.0965323
#define         G2              0.0430785

// Improved Harris corner detector w/o output image
int
ImpHarris(int *Im, int width, int height,
          int **up0, int **vp0, int *cnr_num,
          int Ns, double sigma, double rk,
          double s, int max_corner, double min_dist)
{
   double	r, d;
   double	Ix, Iy;
   double	Ix2, Iy2, Ixy;
   double	gx, gy;
   double	gxx, gyy, gxy;
   double	max_r, r_thd;
   int		corner_num;
   int		x, x0, xx, u0, v0;
   int		uoff, voff, offwidth, offheight;
   int		dblw;
   
   int		*up = new int [max_corner];
   int		*vp = new int [max_corner];
   
   double	*Gw   = new double [(2*Ns+1) * (2*Ns+1)];
   float	*Imx  = new float [width * height];
   float	*Imy  = new float [width * height];
   float	*Imxx = new float [width * height];
   float	*Imyy = new float [width * height];
   float	*Imxy = new float [width * height];
   float	*out  = new float [width * height];
   double	*dp = new double [max_corner];

   // memory allocation check
   if (!up || !vp || !Gw || !Imx || !Imy || !Imxx || !Imyy || !Imxy ||
       !out || !dp)
      return 0;

   (*up0) = up;
   (*vp0) = vp;
   
   make_Gauss_window(Gw, Ns, sigma);
   dblw = width * 2;

   uoff = voff = 2;
   offwidth = width - uoff;
   offheight = height - voff;
   
   for (v0 = voff; v0 < offheight; v0++)
   {
      x = (v0 * width) + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 Ix = G2 * (Im[x + 2] - Im[x - 2])
            + G1 * (Im[x + 1] - Im[x - 1]);
	 Iy = G2 * (Im[x + dblw] - Im[x - dblw])
	    + G1 * (Im[x + width] - Im[x - width]);
	 
	 Imx[x] = Ix;
	 Imy[x] = Iy;
      }
   }

   uoff += 2;
   voff += 2;
   offwidth -= 2;
   offheight -= 2;

   for (v0 = voff; v0 < offheight; v0++)
   {
      x = (v0 * width) + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 gx = Imx[x - 2] + Imx[x - 1] + Imx[x] + Imx[x + 1] + Imx[x + 2];
	 gy = Imy[x - dblw] + Imy[x - width] + Imy[x]
	    + Imy[x + width] + Imy[x + dblw];
	 Imxx[x] = gx * gx;
	 Imyy[x] = gy * gy;
	 Imxy[x] = gx * gy;
      }
   }

   uoff += Ns;
   voff += Ns;
   offwidth -= Ns;
   offheight -= Ns;
   max_r = 0.0;
   
   for (v0 = voff; v0 < offheight; v0++)
   {
      x = v0 * width + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 gxx = gyy = gxy = 0.0;
	 xx = 0;
	 for (int vv = -Ns; vv <= Ns; vv++)
	 {
	    x0 = (v0 + vv) * width + u0;
	    for (int uu = -Ns; uu <= Ns; uu++)
	    {
	       gxx += Gw[xx] * Imxx[x0 + uu];
	       gyy += Gw[xx] * Imyy[x0 + uu];
	       gxy += Gw[xx] * Imxy[x0 + uu];
	       xx++;
	    }
	 }
	 
	 Ix2 = gxx * gxx;
	 Iy2 = gyy * gyy;
	 Ixy = gxy * gxy;
	 r = (Ix2 * Iy2 - sqr(Ixy)) - rk * sqr(Ix2 + Iy2);

	 if (r > max_r)
	    max_r = r;
	 
	 out[x] = r;
      }
   }

   r_thd = max_r * s;

   uoff += 2;
   voff += 2;
   offwidth -= 2;
   offheight -= 2;

   corner_num = 0;
   for (v0 = voff; v0 < offheight; v0++)
   {
      x = v0 * width + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 d = out[x];
	 if (d <= r_thd)  continue;

	 if (   check_ge_upperleft(d, out, x, width)
             && check_gt_lowerright(d, out, x, width))
	 {
	    corner_num = add_list(u0, v0, d, up, vp, dp,
				  corner_num, max_corner, min_dist);
	 }
      }
   }

//   for (int ii = 0; ii < corner_num; ii++)
//      cerr << "d[" << ii << "] = " << dp[ii] << endl;

// delete arrays
   delete [] dp;
   delete [] out;
   delete [] Imxy;
   delete [] Imyy;
   delete [] Imxx;
   delete [] Imy;
   delete [] Imx;
   delete [] Gw;
   
   return (*cnr_num = corner_num);
}

// Harris corner detector
int
ImpHarris(int *Im, int *Imo, int width, int height,
          int **up0, int **vp0, int *cnr_num,
          int Ns, double sigma, double rk,
          double s, int max_corner, double min_dist)
{
   int		x, corner_num;

   // feature detection
   corner_num = ImpHarris(Im, width, height, up0, vp0, cnr_num,
                          Ns, sigma, rk, s, max_corner, min_dist);

   // copy image
   x = 0;
   for (int v = 0; v < height; v++)
      for (int u = 0; u < width; u++)
	 Imo[x++] = Im[x];

   // plot corners
   for (int i = 0; i < corner_num; i++)
   {
      x = (*vp0)[i] * width + (*up0)[i];
      Imo[x - width - 1] = MaxGray;
      Imo[x - width]     = MaxGray;
      Imo[x - width + 1] = MaxGray;
      Imo[x - 1]         = MaxGray;
      Imo[x]             = 0;
      Imo[x + 1]         = MaxGray;
      Imo[x + width - 1] = MaxGray;
      Imo[x + width]     = MaxGray;
      Imo[x + width + 1] = MaxGray;
   }

   return corner_num;
}
