//
// Improved Harris ڥ졼ƥ
//
#include <iostream>
#include <fstream>

using namespace std;

#include <unistd.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>

#include "Harris.h"
#include "imagefile.h"

void
usage(char *command)
{
   cerr << "usage: " << command << " "
	<< "[-f flt_size] "
	<< "[-g sigma ] "
	<< "[-k k_value] "
	<< "[-r rel_ratio] "
	<< "[-n max_cnr_no] "
	<< "[-d min_dist] "
	<< "[-o output.pgm] "
	<< "[-c cornerlist] "
	<< "input.pgm" << endl;
   cerr << "  Input file:  input.pgm" << endl;
   cerr << "  Outputs:     cornerlist, [output.pgm]" << endl;
   cerr << "  Options:" << endl;
   cerr << "        -f flt_size     size of gaussian smoothing filter (5x5)" << endl;
   cerr << "        -g sigma        sigma of gaussian smoothing filter (1)" << endl;
   cerr << "        -k k_value      k value for Harris (0.04)" << endl;
   cerr << "        -r rel_ratio    relative ratio to maximum Harris value >= 0.0 (0.0001)" << endl;
   cerr << "        -n max_cnr_nr   number of detecting corners (200)" << endl;
   cerr << "        -d min_dist     minimum distance between corners (8)" << endl;
   cerr << "        -o output.pgm   output pgm image (none)" << endl;
   cerr << "        -c cnrlist      coordinates list of corners (stderr)" << endl;
}

void
output_corners(char *filename, int *u0, int *v0, int N)
{
   ofstream	fout(filename);
   char         buff[30];

   if (fout.bad())
   {
      cerr << "output_corners(): unable to open " << filename << endl;
      return;
   }

   fout << N << endl;
   for (int i = 0; i < N; i++)
   {
      sprintf(buff, "(%8d,%8d)", u0[i], v0[i]);
      fout << buff << endl;
   }

   fout.close();
}


int
main(int argc, char **argv)
{
   int		*imin, *imout;
   char		im_iname[80], im_oname[80], cnr_name[80], buff[30];
   int		width, height;
   int		fsize;
   double	sigma;
   double	rk;
   double	s;
   double	min_d;
   int		l, num, max_c;
   int		*u0, *v0;
   int		opt;

   // defualts
   fsize = Mask_Size;	// Ⱦʬꡥºݤ 5x5 ΥȤʤ롥
   sigma = Sigma;
   rk = K_value;
   s = Rel_r;
   max_c = Max_Corner;
   min_d = Min_Distance;
   im_iname[0] = im_oname[0] = cnr_name[0] = '\0';
   
   // options
   while ((opt = getopt(argc, argv, "f:g:k:r:n:d:o:c:")) != -1)
   {
      switch (opt)
      {
	case 'f':
	   fsize = atoi(optarg);
	   fsize /= 2;
	   break;
	case 'g':
	   sigma = atof(optarg);
	   break;
	case 'k':
	   rk = atof(optarg);
	   break;
	case 'r':
	   s = atof(optarg);
	   break;
	case 'n':
	   max_c = atoi(optarg);
	   break;
	case 'd':
	   min_d = atof(optarg);
	   break;
	case 'o':
	   strcpy(im_oname, optarg);
	   break;
	case 'c':
	   strcpy(cnr_name, optarg);
	   break;
      }
   }

   if (optind < argc)
      strcpy(im_iname, argv[optind++]);
   else
   {
      usage(argv[0]);
      exit(1);
   }

   if (s < 0.0)
   {
      cerr << "Error: rel_ratio must be a positive or zero" << endl;
      exit(1);
   }

   load_pgm_file(&imin, im_iname, &width, &height);

   if (im_oname[0] == '\0')
      l = ImpHarris(imin, width, height, &u0, &v0, &num,
                    fsize, sigma, rk, s, max_c, min_d);
   else
   {
      imout = new int [width * height];
      l = ImpHarris(imin, imout, width, height, &u0, &v0, &num,
                    fsize, sigma, rk, s, max_c, min_d);
      if (l > 0)
	 save_pgm_file(imout, im_oname, width, height);
   }

   if (l <= 0)
   {
      cout << "Error: cannot detect any corners!" << endl;
      exit(1);
   }
   
   if (cnr_name[0] != '\0')
      output_corners(cnr_name, u0, v0, num);
   else
   {
      cout << num << endl;
      for (int i = 0; i < num; i++)
      {
         sprintf(buff, "(%8d,%8d)", u0[i], v0[i]);
	 cout << buff << endl;
      }
   }
}
